/*
 * vfs.c
 *
 * Copyright 2007, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * scI
 *	@JĂ鉼̊֐폜B
 *	@𒊏oĎB
 *	@opendir()̎dllookup()ΉɂB
 */


#include <sys/config.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <kern/mm.h>
#include <kern/lock.h>
#include <kern/lib_path.h>
#include <kern/device.h>
#include <kern/fs.h>
#include <kern/vfs.h>

#include <kern/test.h>
#include <kern/debug.h>


//#define DEBUG_VFS 1
#ifdef DEBUG_VFS
	#define STATIC
	#define INLINE
#else
	#define STATIC	static
	#define INLINE	inline
#endif


/******************************************************************************************************
 *
 * zt@CVXe
 *
 *  }EgEJgfBNgύXEt@CI[vEt@CsɃGg[ǉAinodeObjvfB
 *  Gg[N̍Ō͍ŏ̃Gg[ɘAB
 * 
 * t@ČE쐬           t@C폜
 * lookup()                           deleteObj()
 * creat()
 *                                   
 * t@CI[v             t@CN[Y
 * open()                             close()
 *                                   
 * zGg[̍쐬             zGg[폜
 * initVent()                         deleteVent()
 *                                   
 * zt@CVXeɐڑ       zt@CVXeؒf
 * addVent()
 *                                   
 * N                           AN
 * linkVent()                         unlinkVent()
 *******************************************************************************************************/

/*
 * zGg[
 * Note!
 *  fsinodëʒu̓\Pbg\̂ƃNĂ̂ŒӁB
 */
typedef struct VENTRY{
	FS				*fs;		// t@CVXe
	void			*inodeObj;	// inode object |C^
	struct VENTRY	*next;		// EhNXglNXg
	struct VENTRY	*prev;		// EhNXgvr[
	struct VENTRY	*parent;	// eGg[
	struct VENTRY	*child;		// ̃m[h̃gbvGg[
	struct VENTRY	*mount;		// }EgGg[
	struct VENTRY	*mounted;	// }EgGg[
	int				type;		// t@C^Cv
	uchar			linkCount;	// I[vA}EgAJgfBNgJEgAbv
	uchar			din;		// foCXinode
	uchar			del;		// 폜vTCiYES or NOj
	uchar			len;		// Name size.
	char			name[0];	// pXl[
} VENTRY;

//================================== PRIVATE ============================================

static FS fsInfoTop;							// t@CVXeo^Ngbv
static char rootEntArea[sizeof(VENTRY) + 1];
static VENTRY *rootEnt = (VENTRY*)&rootEntArea;	// [gGg[

//------------------ zt@Cb\h̍XVEQƃbN ------------

static  MODIFY_WAIT_LOCK fsModifyLock;			// zt@CVXeEt@CVXeR[bN

/*
 * zt@CVXe̎QƊJn
 */
STATIC INLINE void refVfs()
{
	modifyWaitRef(&fsModifyLock);
}

/*
 * zt@CVXe̎QƏI
 */
STATIC INLINE void refEndVfs()
{
	modifyWaitRefEnd(&fsModifyLock);
}

/*
 * zt@CVXe̍XVJn
 */
STATIC INLINE void modifyVfs()
{
	modifyWaitLock(&fsModifyLock);
}

/*
 * zt@CVXe̍XVI
 */
STATIC INLINE void modifyEndVfs()
{
	modifyWaitUnlock(&fsModifyLock);
}

//------------------ pX擾E -------------------------------

/*
 * copyPathBack()p֐
 */
STATIC int isRoot(const VENTRY *vent)
{
	return (vent == rootEnt) ? YES : NO;
}

STATIC int isMount(const VENTRY *vent)
{
	return (vent->mount != NULL) ? YES : NO;
}

STATIC int isParent(const VENTRY *vent)
{
	return (vent->inodeObj != NULL) ? YES : NO;
}

/*
 *sQtsċAt
 * pXăobt@ɃRs[B
 * return : Rs[TCY
 */
STATIC int copyPathBack(
	const VENTRY *vent,					// XgGg[
	const size_t size,					// Rs[őTCY
	int (*condition)(const VENTRY *),	// k
	char *m_path)						// Rs[obt@
{
	size_t cpSize;

	if (condition(vent) == YES) {
		cpSize = (size < vent->len) ? size : vent->len;
		memcpy(m_path, vent->name, cpSize);
		return cpSize;
	}
	else{
		int len;

		len = copyPathBack(vent->parent, size, condition, m_path);
		if (size <= len) {
			return len;
		}
		m_path[len++] = '/';
		cpSize = (size - len < vent->len) ? size - len : vent->len;
		memcpy(&m_path[len], vent->name, cpSize);

		return len + cpSize;
	}
}

/*
 *sQtsċAt
 * pXefBNg܂şăobt@ɃRs[B
 * ̃Gg[[g܂̓}Eg[gGg[Ȃ玩Gg[̃pXݒ肷B
 * return : Rs[TCY
 */
STATIC int copyParentPath(
	const VENTRY *vent,		// XgGg[
	const size_t size,		// Rs[őTCY
	char *m_path)			// Rs[obt@
{
	int len = copyPathBack(vent->parent, size, &isParent, m_path);

	if (vent != vent->parent) {
		size_t cpSize;

		if (size <= len) {
			return len;
		}
		m_path[len++] = '/';
		cpSize = (size - len < vent->len) ? size - len : vent->len;
		memcpy(&m_path[len], vent->name, cpSize);
		len += cpSize;
	}
	
	return len;
}

/*
 *sQt
 * eGg[擾
 * return : eGg[
 */
STATIC VENTRY *getParentVent(const VENTRY *vent)
{
	if (vent->parent->inodeObj != NULL) {
		return vent->parent;
	}
	else {
		return getParentVent(vent->parent);
	}
}

/*
 * pathgbvfBNg𓾂B
 * parameters : pXAhX|C^
 * return : JgzfBNgGg[
 */
STATIC VENTRY *getTopVdir(const char *path, const char **o_path)
{
	VENTRY *vent;
	char *p = (char*)path;

	if(*p == '/'){
		vent = rootEnt;
		while(*++p == '/');
		*o_path = p;
	}
	else{
		vent = getCurrentDir();
		*o_path = p;
	}

	return (vent->mount != NULL) ? vent->mount : vent;
}

//------------------ t@C̍쐬E폜E -------------------------------

/*
 *sXt
 * zGg[IuWFNgit@CEfBNgj폜B
 */
STATIC void deleteObjFromVent(
	const VENTRY *vent,		// 폜Gg[
	const int type)			// 폜Gg[^Cv
{
	char path[PATH_MAX + 1];
	VENTRY *parent;
	int len;

	parent = getParentVent(vent);
	len = copyParentPath(vent, PATH_MAX + 1, path);
	if (PATH_MAX < len){
		return;
	}
	path[len] = '\0';
	parent->fs->delete(path, parent->din, parent->inodeObj, type);
}

/*
 *sXt
 * IuWFNg̍폜tO𗧂Ă
 */
STATIC void unlinkObj(VENTRY *m_vent)
{
	m_vent->del = YES;
}

//------------------ zGg[̍쐬E폜 ---------------------

/*
 *sXt
 * Gg[ݒ肷B
 */
STATIC void initVent(
	const int din,
	FS *fs,
	void *inodeObj,
	const int type,
	VENTRY *m_vent)
{
	m_vent->din = din;
	m_vent->fs = fs;
	m_vent->inodeObj = inodeObj;
	m_vent->type = type;
	m_vent->del = NO;
}

/*
 *sXt
 * zGg[폜B
 * return : error number
 */
STATIC int deleteVent(VENTRY *m_vent)		// 폜Gg[
{
	if (0 < m_vent->linkCount) {
		return -EBUSY;
	}
	if (m_vent->child != NULL) {
		return -EBUSY;
	}
	if (m_vent->mount != NULL) {
		return -EBUSY;
	}
	if (m_vent->mounted != NULL) {
		VENTRY *mounted = m_vent->mounted;
		int error;

		m_vent->fs->umount(m_vent->din, m_vent->inodeObj);

		mounted->mount = NULL;
		error = deleteVent(mounted);
		if (error != NOERR) {
			return error;
		}
	}

	//t@CN[Y
	if (m_vent->inodeObj != NULL) {
		m_vent->fs->close(m_vent->din, m_vent->inodeObj);
	}
	
	//t@C폜
	if (m_vent->del == YES) {
		deleteObjFromVent(m_vent, m_vent->type);
	}

	//eGg[̎qN߂
	if (m_vent->parent != NULL) {
		VENTRY *parent = m_vent->parent;

		if (m_vent->next == m_vent) {
			parent->child = NULL;
		}
		else {
			if (parent->child == m_vent) {
				parent->child = m_vent->next;
			}
			m_vent->prev->next = m_vent->next;
			m_vent->next->prev = m_vent->prev;
		}
	}

	kfree(m_vent);

	return NOERR;
}

//------------------ zt@CVXeɐڑEؒfE ---------------

/*
 *sQtst
 * m[hGg[TB
 * return : error number
 */
STATIC int searchVentNode(
	const char *path,
	VENTRY *parent,			// eGg[
	VENTRY **o_entry)		// Gg[
{
	VENTRY *entry;

	if (cmpPath(".", path, 1) == 0){
		*o_entry = parent;
		return NOERR;
	}
	if (cmpPath("..", path, 2) == 0){
		*o_entry = parent->parent;
		return NOERR;
	}

	// m[hNT
	if (parent->child == NULL){
		return -ENOENT;
	}
	entry = parent->child;
	while (cmpPath(entry->name, path, entry->len) != 0) {
		entry = entry->next;
		if (entry == parent->child){
			return -ENOENT;
		}
	}

	*o_entry = (entry->mount == NULL) ? entry : entry->mount;

	return NOERR;
}

/*
 *sXtst
 * m[hŃGg[쐬EǉB
 * return : error number
 */
STATIC int addVentNode(
	const char *path,
	VENTRY *m_parent,
	VENTRY **o_vent)
{
	VENTRY *vent;
	int error;

	error = searchVentNode(path, m_parent, &vent);
	switch (error) {
	case NOERR:
		// ".."̏ꍇ͍폜
		if (cmpPath("..", path, 2) == 0) {
			deleteVent(m_parent);
		}
		break;
	case -ENOENT:
		vent = kmalloc(sizeof(*vent) + pathLen(path));
		if (vent == NULL) {
			return -ENOMEM;
		}
		memset(vent, 0, sizeof(*vent));

		// Nɒǉ
		if (m_parent->child == NULL){
			vent->prev = vent->next = vent;
			m_parent->child = vent;
		}
		else{
			VENTRY *child = m_parent->child;
			vent->prev = child;
			vent->next = child->next;
			child->next->prev = vent;
			child->next = vent;
		}
		vent->parent = m_parent;
		vent->child = NULL;
		vent->len = pathLen(path);
		memcpy(vent->name, path, vent->len);

		break;
	default:
		return error;
	}

	*o_vent = vent;

	return NOERR;
}

/*
 *sXtst
 *sӁt̐eGg[͓t@CVXełȂ΂ȂȂB
 * Gg[pXɒǉB
 * return : error number
 */
STATIC int addVent(
	const char *i_path,		//ẽpX
	VENTRY *m_parent,		//eGg[
	VENTRY **o_vent)
{
	const char *path = i_path;
	VENTRY *parent;
	VENTRY *vent;
	int error;

	ASSERT(*path != NULL);

	parent = m_parent;
	for(;;){
		//Gg[̒ǉ
		error = addVentNode(path, parent, &vent);
		if (error != NOERR) {
			return error;
		}

		path = getNextPath(path);
		if (*path == '\0'){
			break;
		}
		else {
			initVent(parent->din, parent->fs, NULL, S_IFDIR, vent);
		}

		parent = vent;
	}

	*o_vent = vent;

	return error;
}

/*
 *sQtst
 * pXGg[B
 * return : error number
 */
STATIC int searchVent(
	const char *i_path,			// pX
	const VENTRY *i_vent,		// JninodeIuWFNg
	const char **o_parentPath,	// inodeIuWFNgefBNgpX
	VENTRY **o_parentVent,		// inodeIuWFNgefBNgzGg[
	VENTRY **o_vent)			// Ō̉zGg[
{
	const char *path;
	VENTRY *vent;
	int error;

	path = i_path;
	vent = (VENTRY*) i_vent;
	*o_vent = vent;
	for(;;) {
		// Gg[𔻒肷
		if ((vent->inodeObj != NULL) || (vent->mounted != NULL )) {
			*o_parentPath = path;
			*o_parentVent = vent;
			error = NOERR;
		}
		else {
			error = -ENOENT;
		}

		if (*path == '\0') {
			break;
		}
		
		if (vent->type != S_IFDIR) {
			return -ENOTDIR;
		}

		// qGg[擾
		error = searchVentNode(path, vent, &vent);
		if (error != NOERR) {
			return error;
		}

		*o_vent = vent;
		path = getNextPath(path);
	}

	return error;
}

//------------------ zGg[̃NEAN -------------

/*
 *sQt
 * zGg[I[v
 */
void linkVent(void *i_vent)
{
	VENTRY *m_vent = i_vent;

	ASSERT(m_vent->linkCount < _POSIX_OPEN_MAX);

	++m_vent->linkCount;
}

/*
 *sXt
 *sťĂяo͕KP݂̂ɂȂ邱ƁBiN[Y_uČĂяoȂƁj
 * zGg[N[Y
 */
STATIC void unlinkVent(VENTRY *m_vent)
{
	ASSERT(0 < m_vent->linkCount);

	--m_vent->linkCount;

	if (m_vent->linkCount == 0){
		VENTRY *vent;
		VENTRY *parent;

		// zGg[kč폜B
		for(vent = m_vent, parent = vent->parent; deleteVent(vent) == NOERR;) {
			vent = parent;
			parent = vent->parent;
		}
	}
}

//------------------ V{bNNΉpX̌ ----------------------

enum {
	SYMLINK_LOOP_MAX = 10,		// V{bNNőێ
};

/*
 * V{bNN[v`FbNpe[u
 */
typedef struct {
	int count;							// V{bNNJEg
	int din[SYMLINK_LOOP_MAX];			// foCXinode
	uint inodeNum[SYMLINK_LOOP_MAX];	// indoeԍ
} SYMLINK_LOOP;

/*
 * V{bNN[ve[ȕ
 */
STATIC void initSymlinkLoop(SYMLINK_LOOP *m_symlinkLoop)
{
	m_symlinkLoop->count = 0;
}

/*
 * V{bNN[v`FbN
 * return : error number
 */
STATIC int symlinkLoop(
	const int din,
	const uint inodeNum,
	SYMLINK_LOOP *m_symlinkLoop)
{
	int i;

	ASSERT(m_symlinkLoop->count <= SYMLINK_LOOP_MAX);
	
	if (m_symlinkLoop->count == SYMLINK_LOOP_MAX) {
		return -ELOOP;
	}
	for (i = 0; i < m_symlinkLoop->count; ++i) {
		if ((m_symlinkLoop->din[i] == din) && (m_symlinkLoop->inodeNum[i] == inodeNum)) {
			return -ELOOP;
		}
	}
	
	m_symlinkLoop->din[m_symlinkLoop->count] = din;
	m_symlinkLoop->inodeNum[m_symlinkLoop->count] = inodeNum;
	++m_symlinkLoop->count;
	
	return NOERR;
}

/*
 *sQt
 * V{bNÑpXZbgB
 * return : error number
 */
STATIC int setSymlinkPath(
	const int din,
	const FS *fs,
	const uint inodeNum,
	const char *path,
	const char *symlinkPath,
	char *m_path,					// pX PATH_MAX + 1 ł邱
	SYMLINK_LOOP *m_symlinkLoop)
{
	char *pathBuf;
	const char *nextPath;
	void *inodeObj;
	int lenPrev;
	int len;
	int dummy;
	int error;

	error = symlinkLoop(din, inodeNum, m_symlinkLoop);
	if (error != NOERR) {
		return error;
	}

	pathBuf = m_path;

	// V{bNNpX
	error = fs->open(din, inodeNum, O_RDONLY, &inodeObj, &dummy);
	if (error != NOERR) {
		return error;
	}
	len = fs->read(din, inodeObj, pathBuf, PATH_MAX, 0);
	if (fs->close(din, inodeObj) != NOERR) {
		ASSERT(0);
	}
	if (len < 0) {
		return len;
	}
	
	if (*pathBuf == '/') {
		// ΃pXȂ̂őOɃpX͕tȂ
		lenPrev = 0;
	}
	else {
		// V{bNNO܂ł̃pXt
		lenPrev = symlinkPath - path;
		if (PATH_MAX < lenPrev + len) {
			return -ENAMETOOLONG;
		}
		memmove(pathBuf + lenPrev, pathBuf, len);
		memcpy(pathBuf, path, lenPrev);
	}
	pathBuf += lenPrev + len;

	// V{bNÑpX
	nextPath = getNextPath(symlinkPath);
	if (*nextPath != '\0') {
		int lenNext;

		nextPath -= 1;			// '/'܂Ŗ߂
		lenNext = strnlen(nextPath, PATH_MAX + 1);
		if (PATH_MAX < lenPrev + len + lenNext) {
			return -ENAMETOOLONG;
		}
		memcpy(pathBuf, nextPath, lenNext);
		pathBuf += lenNext;
	}
	*pathBuf = '\0';

	return NOERR;
}

/*
 *sQt
 * t@C
 * return : error number or SYMLINK
 */
STATIC int lookupObj(
	const char *i_path, 	// pX
	const int din,
	const FS *fs,
	void *i_inodeObj,
	char *m_path,			// V{bNNɂcpXpXobt@
	SYMLINK_LOOP *m_symlinkLoop,
	uint *o_inodeNum)		// inodeԍo
{
	uint inodeNum;
	const char *symlinkPath;
	int error;

	if(*i_path=='\0'){
		return -ENOENT;
	}

	error = fs->lookup(i_path, din, i_inodeObj, &inodeNum, &symlinkPath);
	switch (error) {
	case NOERR:
		break;
	case SYMLINK:		// V{bNN
		// pX̐ݒ
		error = setSymlinkPath(din, fs, inodeNum, i_path, symlinkPath, m_path, m_symlinkLoop);
		if (error != NOERR) {
			return error;
		}
		return SYMLINK;
	default:
		return error;
	}

	*o_inodeNum = inodeNum;

	return error;
}

/*
 *sQt
 * t@C
 * V{bNN
 * return : error number or SYMLINK
 */
STATIC int lookupObjSymlink(
	const char *i_path, 	// pX
	const int din,
	const FS *fs,
	void *i_inodeObj,
	char *m_path,			// V{bNNɂcpXpXobt@
	SYMLINK_LOOP *m_symlinkLoop,
	uint *o_inodeNum)		// inodeԍo
{
	uint inodeNum;
	const char *symlinkPath;
	int error;

	if(*i_path=='\0'){
		return -ENOENT;
	}

	error = fs->lookup(i_path, din, i_inodeObj, &inodeNum, &symlinkPath);
	switch (error) {
	case NOERR:
		break;
	case SYMLINK:		// V{bNN
		// pX̍Ō̃Gg[Ȃ琳I
		if (*getNextPath(symlinkPath) == '\0') {
			error = NOERR;
		}
		else {
			error = setSymlinkPath(din, fs, inodeNum, i_path, symlinkPath, m_path, m_symlinkLoop);
			if (error != NOERR) {
				return error;
			}
			return SYMLINK;
		}
		break;
	default:
		return error;
	}

	*o_inodeNum = inodeNum;

	return error;
}

/*
 * lookupPathpo̓p[^
 */
typedef struct {
	VENTRY *vent;				// zGg[
	int din;					// foCXinode
	FS *fs;						// t@CVXe\
	uint inodeNum;				// inodeԍ
	const char *parentPath;		// efBNgpX
	VENTRY *parentVent;			// efBNginodeIuWFNg
} LOOKUP_OUT;

enum {
	EXIST_VENT = 1,		// zGg[
};

/*
 *sQt
 * zGg[Ǝt@CTAV{bNNΉB
 * łɉzGg[΃AEgp[^ɉzGg[ZbgEXIST_VENTԂB
 */
STATIC int lookupPath(
	const char *i_path,
	int lookupFunc(const char *, const int, const FS *, void *, char *, SYMLINK_LOOP *, uint *),
	LOOKUP_OUT *m_lookupOut)
{
	const char *path;
	const char *parentPath;
	VENTRY *parentVent;
	char linkPath[PATH_MAX + 1];
	VENTRY *vent;
	SYMLINK_LOOP symlinkLoop;
	uint inodeNum;
	int error;

	initSymlinkLoop(&symlinkLoop);

	//JgfBNg擾B
	vent = getTopVdir(i_path, &path);

	for (;;) {
		error = searchVent(path, vent, &parentPath, &parentVent, &vent);
		if (error == NOERR) {				// zt@Cɂ
			m_lookupOut->vent = vent;
			return EXIST_VENT;
		}
		else if (error == -ENOENT) {		// zt@Cɖ
			error = lookupFunc(parentPath, parentVent->din, parentVent->fs, parentVent->inodeObj, linkPath, &symlinkLoop, &inodeNum);
			if (error == NOERR) {
				// m_lookupOutݒ
				break;
			}
			else if (error == SYMLINK) {	// V{bNN
				path = linkPath;
				if (*path == '/') {
					// ΃pX̏ꍇ[gzGg[擾
					vent = getTopVdir(path, &path);
				}
				// ēxsearchVent()
			}
			else if (error == -ENOENT){
				// m_lookupOutݒ
				break;
			}
			else {
				return error;
			}
		}
		else {
			return error;
		}
	}

	m_lookupOut->din = parentVent->din;
	m_lookupOut->fs = parentVent->fs;
	m_lookupOut->inodeNum = inodeNum;
	m_lookupOut->parentPath = parentPath;
	m_lookupOut->parentVent = parentVent;

	return error;
}

//------------------ [`̃bp[ ----------------------

/*
 *sQt
 * zt@CɂΉzGg[𖳂Ύt@CI[vĉzt@C쐬B
 * return : error number
 */
STATIC int getVent(
	const char *i_path,
	const int flag,			// opentO
	int lookupFunc(const char *, const int, const FS *, void *, char *, SYMLINK_LOOP *, uint *),
	VENTRY **o_vent)		// zinode
{
	LOOKUP_OUT lookupOut;
	int error;

	error = lookupPath(i_path, lookupFunc, &lookupOut);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		*o_vent = lookupOut.vent;
		error = NOERR;
		break;
	case NOERR: {		// zt@Cɖ
		void *inodeObj;
		int type;
		VENTRY *vent;

		error = lookupOut.fs->open(lookupOut.din, lookupOut.inodeNum, flag, &inodeObj ,&type);
		if (error != NOERR){
			break;
		}
		vent = kmalloc(sizeof(*vent));
		if (vent == NULL) {
			lookupOut.fs->close(lookupOut.din, inodeObj);
			error = -ENOENT;
			break;
		}
		memset(vent, 0, sizeof(*vent));
		initVent(lookupOut.din, lookupOut.fs, inodeObj, type, vent);
		*o_vent = vent;
		break;
	}
	default:
		break;
	}

	return error;
}

//------------------ t@CVXẽ}Eg -------------

/*
 *sQt
 * }EgfBNg̎擾
 * return : zGg[
 */
STATIC INLINE VENTRY *getMountedVent(VENTRY *vent)
{
	return (vent->mounted != NULL) ? vent->mounted : vent;
}

/*
 * struct statfsɊi[B
 * return : error number
 */
STATIC int getFsstat(VENTRY *i_vent, struct statfs *statfs)
{
	VENTRY *vent;
	int error;
	DEV_STAT devstat;

	vent = (i_vent->mount == NULL) ? i_vent : i_vent->mount;

	if ((error = vent->fs->statfs(vent->din, statfs)) < 0){
		return error;
	}
	get_devstat(vent->din,&devstat);

	if (devstat.prt_num > 0){
		sprintk(statfs->f_mntfromname, "/dev/%s%d", devstat.name, devstat.prt_num);
	}
	else{
		sprintk(statfs->f_mntfromname, "/dev/%s", devstat.name);
	}

	copyPathBack(vent, MNAMELEN, &isMount, statfs->f_mntonname);

	return NOERR;
}

/*
 *sQtsċAt
 * fBNg}EgGg[̌B
 * return }Eg
 */
STATIC int searchMountDir(VENTRY *vent, struct statfs *m_statfs, uint lenStatfs)
{
	int rest, count;
	VENTRY *vent_begin;

	if (lenStatfs <= 0){
		return 0;
	}
	vent_begin = vent;
	count = lenStatfs;

	do{
		// }Eg
		if (vent->mount != NULL){
			while (vent->mount != NULL){
				vent = vent->mount;
			}
			if (m_statfs != NULL){
				if (getFsstat(vent, m_statfs) == NOERR){
					m_statfs += 1;
					count -= 1;
				}
			}
		}

		if (vent->child != NULL){
			rest = searchMountDir(vent->child, m_statfs, count);
			m_statfs += rest;
			count -= rest;
		}
	}while ((vent = vent->next) != vent_begin);

	return lenStatfs - count;
}

STATIC void umountFs(VENTRY *ventry)
{
	VENTRY *vent = ventry;

	do{
		if (vent->child != NULL){
			umountFs(vent->child);
		}
		if (vent->mount != NULL){
			vent->mount->fs->umount(vent->mount->din, vent->mount->inodeObj);
		}
	}while ((vent = vent->next) != ventry);
}

/*
 *sXt
 * }EgGg[A}Eg
 * return : error number
 */
STATIC int umountFromVfs(
	VENTRY *vent)		// }EgGg[
{
	VENTRY *mountVent;

	if (vent->mount == NULL){
		return -ENOENT;
	}
	mountVent = vent->mount;

	if (mountVent->child != NULL) {
		return -EBUSY;
	}

	return NOERR;
}

/*
 * zt@CVXeă}EgĂt@CVXeA}EgB
 */
void umountAllFs()
{
	umountFs(rootEnt);
}

/*
 * return : error number
 */
STATIC int chmod(
	const VENTRY *vent,
	const mode_t mode)
{
	PROC *proc = get_current_task();
	struct stat stat;
	int error;

	// ̃`FbN
	error = vent->fs->stat(vent->din, vent->inodeObj, &stat);
	if (error != NOERR) {
		return error;
	}
	if ((proc->uid != 0) && (proc->uid != stat.st_uid)) {
		return -EPERM;
	}

	return vent->fs->chattr(vent->din, vent->inodeObj, mode, stat.st_uid, stat.st_gid, stat.st_atime, stat.st_mtime);
}

/*
 * return : error number
 */
STATIC int chuid(
	const VENTRY *vent,
	const uid_t owner)
{
	PROC *proc = get_current_task();
	struct stat stat;
	int error;

	ASSERT(owner != -1);

	// ̃`FbN
	if (proc->uid != 0) {
		return -EPERM;
	}

	error = vent->fs->stat(vent->din, vent->inodeObj, &stat);
	if (error != NOERR) {
		return error;
	}

	return vent->fs->chattr(vent->din, vent->inodeObj, stat.st_mode, owner, stat.st_gid, stat.st_atime, stat.st_mtime);
}

/*
 * return : error number
 */
STATIC int chgrp(
	const VENTRY *vent,
	const gid_t group)
{
	PROC *proc = get_current_task();
	struct stat stat;
	int error;

	ASSERT(group != -1);

	// ̃`FbN
	error = vent->fs->stat(vent->din, vent->inodeObj, &stat);
	if (error != NOERR) {
		return error;
	}
	if (((proc->uid != 0) && (proc->uid != stat.st_uid)) ||
	    (group != proc->gid)) {
		return -EPERM;
	}

	return vent->fs->chattr(vent->din, vent->inodeObj, stat.st_mode, stat.st_uid, group, stat.st_atime, stat.st_mtime);
}

/*
 * return : error number
 */
STATIC inline int chutime(
	const VENTRY *vent,
	const uint atime,
	const uint mtime)
{
	PROC *proc = get_current_task();
	struct stat stat;
	int error;

	// ̃`FbN
	error = vent->fs->stat(vent->din, vent->inodeObj, &stat);
	if (error != NOERR) {
		return error;
	}
	if ((proc->uid != 0) && (proc->uid != stat.st_uid)) {
		return -EPERM;
	}

	return vent->fs->chattr(vent->din, vent->inodeObj, stat.st_mode, stat.st_uid, stat.st_gid, atime, mtime);
}

//------------------ t@CVXe̓o^EE폜 -------------------

/*
 * t@CVXeo^B
 * return : 0 or Error=-1
 */
int regist_fs(FS *info)
{
	FS *p;

	p = &fsInfoTop;
	for (; p->next != NULL; p = p->next) {
		switch (strcmp(info->name, p->next->name)) {
		case 0:
			return -EEXIST;
		case -1:
			goto EXIT;
		}
	}
EXIT:
	info->next = p->next;
	p->next = info;

	return 0;
}

/*
 * o^ꂽt@CVXeTB
 * return : FS address or NULL
 */
FS *search_fs(const char *name)
{
	FS *p;

	for (p = fsInfoTop.next; p != NULL; p = p->next) {
		switch (strcmp(name, p->name)) {
		case 0:
			return p;
		case -1:
			return NULL;
		}
	}

	return NULL;
}

//------------------ zEt@C -------------------

/*
 *sXt
 * w̃t@CVXẽ[ginode}EgB
 * return : error number
 */
int mountToVfs(
	FS *fs,
	int din,
	void *rootInodeObj,
	void *m_dstVent)
{
	VENTRY *dstVent = m_dstVent;
	VENTRY *mountVent;

	// zt@CVXe̍XVJn
	modifyVfs();

	// łɃ}EgĂ邩H
	if (dstVent->mount != NULL){
		return -EBUSY;
	}

	// }EgGg[̍쐬
	mountVent = kmalloc(sizeof(VENTRY) + dstVent->len);
	if (mountVent == NULL){
		return -ENOMEM;
	}
	memset(mountVent, 0, sizeof(VENTRY));

	// }EgGg[̏
	mountVent->next		= dstVent->next;
	mountVent->prev		= dstVent->prev;
	mountVent->inodeObj	= rootInodeObj;
	mountVent->parent	= dstVent->parent;
	mountVent->mounted	= dstVent;
	mountVent->fs		= fs;
	mountVent->din		= din;
	mountVent->type		= S_IFDIR;
	mountVent->len		= dstVent->len;
	memcpy(mountVent->name, dstVent->name, dstVent->len);
	linkVent(mountVent);

	// }Eg
	dstVent->mount = mountVent;

	// zt@CVXe̍XVI
	modifyEndVfs();

	return NOERR;
}

/*
 * return : error number
 */
int umountPath(const char *i_path)
{
	LOOKUP_OUT lookupOut;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = lookupPath(i_path, lookupObj, &lookupOut);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		if (lookupOut.vent->type == S_IFDIR){
			error = umountFromVfs(lookupOut.vent);
			unlinkVent(lookupOut.vent);
		}
		else {
			error = -ENOENT;
		}
		break;
	case NOERR: {		// zt@Cɖ
		error = -ENOENT;
		break;
	}
	default:
		break;
	}

	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 *sXt
 * t@C̍쐬EI[v
 * return : error number
 */
int creatObj(
	const char *i_path,
	const mode_t mode)
{
	const char *path;
	const char *parentPath;
	VENTRY *parentVent;
	VENTRY *vent;
	uint inodeNum;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	//JgfBNg擾B
	vent = getTopVdir(i_path, &path);

	error = searchVent(path, vent, &parentPath, &parentVent, &vent);
	switch (error){
	case NOERR:		// zt@Cɂ
		error = -EEXIST;
		break;
	case -ENOENT:	// zt@Cɖ
		error = parentVent->fs->creat(parentPath, parentVent->din, parentVent->inodeObj, mode, &inodeNum);
		break;
	default:
		break;
	}

	// zt@CVXe̍XVI
	modifyEndVfs();

	return NOERR;
}

/*
 * return : error number
 */
int mkdirPath(
	const char *i_path,
	const mode_t mode)
{
	LOOKUP_OUT lookupOut;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = lookupPath(i_path, lookupObj, &lookupOut);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		//
	case NOERR:			// t@Cɂ
		error = -EEXIST;
		break;
	case -ENOENT:			// t@C
		error = lookupOut.fs->mkdir(lookupOut.parentPath, lookupOut.din, lookupOut.parentVent->inodeObj, mode);
		break;
	default:
		break;
	}

	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 *sXt
 * V{bNN̍쐬
 * return : error number
 */
int creatSymlink(
	const char *i_path,		// 쐬V{bNNpX
	const mode_t mode,		// V{bNNt@C[h
	char *linkPath)	// NpX
{
	const char *path;
	const char *parentPath;
	VENTRY *parentVent;
	VENTRY *vent;
	uint inodeNum;
	void *inodeObj;
	int len;
	VENTRY *dummy1;
	int dummy2;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	//JgfBNg擾B
	vent = getTopVdir(i_path, &path);

	error = searchVent(path, vent, &parentPath, &parentVent, &dummy1);
	switch (error){
	case NOERR:		// zt@Cɂ
		error = -EEXIST;
		goto ERR;
	case -ENOENT:	// zt@Cɖ
		break;
	default:
		goto ERR;
	}

	error = parentVent->fs->creat(parentPath, parentVent->din, parentVent->inodeObj, mode, &inodeNum);
	if (error != NOERR) {
		goto ERR;
	}
	error = parentVent->fs->open(parentVent->din, inodeNum, O_WRONLY, &inodeObj ,&dummy2);
	if (error != NOERR){
		goto ERR;
	}
	len = strnlen(linkPath, PATH_MAX + 1);
	error = parentVent->fs->write(parentVent->din, inodeObj, linkPath, len, 0);
	if (0 < error) {
		error = NOERR;
	}
	parentVent->fs->close(parentVent->din, inodeObj);
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * zGg[I[v
 * zGg[ɂȂ΍쐬ăI[vB
 * return : error number
 */
int openPath(
	const char *i_path,
	const int flag,
	void **o_vent)
{
	LOOKUP_OUT lookupOut;
	VENTRY *vent;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = lookupPath(i_path, lookupObj, &lookupOut);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		vent = lookupOut.vent;
		error = NOERR;
		break;
	case NOERR: {		// zt@Cɖ
		void *inodeObj;
		int type;

		error = lookupOut.fs->open(lookupOut.din, lookupOut.inodeNum, flag, &inodeObj ,&type);
		if (error != NOERR){
			goto ERR;
		}

		// zfBNg̍쐬
		error = addVent(lookupOut.parentPath, lookupOut.parentVent, &vent);
		if (error != NOERR){
			lookupOut.fs->close(lookupOut.din, inodeObj);
			goto ERR;
		}

		initVent(lookupOut.din, lookupOut.fs, inodeObj, type, vent);
		break;
	}
	default:
		goto ERR;
	}

	linkVent(vent);
	*o_vent = vent;
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * t@CI[vĉzGg[ŕԂB
 * return : error number
 */
int openObj(
	const char *i_path,
	const int flag,			// opentO
	void **o_vent)			// zinode
{
	int error;
	
	// zt@CVXe̎QƊJn
	refVfs();

	error = getVent(i_path, flag, lookupObj, (VENTRY**) o_vent);

	// zt@CVXe̎QƏI
	refEndVfs();

	return error;
}

/*
 * return : error number
 */
int opendirVent(
	void *i_vent,
	uint *o_block,
	int *o_index)
{
	VENTRY *vent = i_vent;
	int error;

	// zt@CVXe̎QƊJn
	refVfs();

	error = vent->fs->opendir(vent->din, vent->inodeObj, o_block, o_index);

	// zt@CVXe̎QƏI
	refEndVfs();
	
	return error;
}

/*
 * return : error number
 */
int readdirVent(
	void *i_vent,
	uint *m_block,
	int *m_index,
	char *m_name)
{
	VENTRY *vent = i_vent;

	return vent->fs->readdir(vent->din, vent->inodeObj, m_block, m_index, m_name);
}

/*
 * zGg[N[YB
 * return : error number
 */
void closeVent(void *i_vent)
{
	VENTRY *vent = i_vent;

	ASSERT(vent->parent != NULL);

	// zt@CVXe̍XVJn
	modifyVfs();

	unlinkVent(vent);

	// zt@CVXe̍XVI
	modifyEndVfs();
}

/*
 * getVent()Ŏ擾zGg[폜B
 */
void releaseVent(void *i_vent)
{
	VENTRY *vent = i_vent;

	if (vent->parent == NULL) {
		vent->fs->close(vent->din, vent->inodeObj);
		kfree(vent);
	}
}

/*
 * t@C폜B
 * return : error number
 */
int deleteObj(
	const char *i_path,
	const int type)
{
	LOOKUP_OUT lookupOut;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = lookupPath(i_path, lookupObjSymlink, &lookupOut);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		unlinkObj(lookupOut.vent);
		error = -EBUSY;
		break;
	case NOERR:			// zt@Cɖ
		error = lookupOut.fs->delete(lookupOut.parentPath, lookupOut.din, lookupOut.parentVent->inodeObj, type);
		break;
	default:
		break;
	}

	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int renamePath(
	const char *oldPath,
	const char *newPath)
{
	LOOKUP_OUT lookupOutOld;
	LOOKUP_OUT lookupOutNew;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	// renamet@C
	error = lookupPath(oldPath, lookupObjSymlink, &lookupOutOld);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		error = -EBUSY;
		goto ERR;
	case NOERR:			// t@Cɂ
		break;
	default:
		goto ERR;
	}

	// renamet@C
	error = lookupPath(newPath, lookupObjSymlink, &lookupOutNew);
	switch (error) {
	case EXIST_VENT:	// zt@Cɂ
		//
	case NOERR:			// t@Cɂ
		//
	case -ENOENT:		// t@Cɖ
		break;
	default:
		goto ERR;
	}

	if (lookupOutOld.din == lookupOutNew.din) {
		error = lookupOutOld.fs->rename(lookupOutOld.din, 
			lookupOutOld.parentVent->inodeObj, lookupOutOld.parentPath, 
			lookupOutNew.parentVent->inodeObj, lookupOutNew.parentPath);
	}
	else {
		error = -EXDEV;
	}
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * zGg[̃f[^ǂ
 * return : read bytes or error number
 */
int readVent(
	void *i_vent,
	const size_t size,		// readTCY
	const size_t offset,	// readJnItZbg
	void *m_buf)			// readobt@
{
	VENTRY *vent = i_vent;

	return vent->fs->read(vent->din, vent->inodeObj, m_buf, size, offset);
}

/*
 * zGg[̃f[^ɏ
 * return : read bytes or error number
 */
int writeVent(
	void *i_vent,
	const size_t size,		// writeTCY
	const size_t offset,	// writeJnItZbg
	void *m_buf)			// writeobt@
{
	VENTRY *vent = i_vent;

	return vent->fs->write(vent->din, vent->inodeObj, m_buf, size, offset);
}

/*
 * pXt@CXe[g擾
 * return : error number
 */
int statPath(
	const char *i_path,
	struct stat *m_stat)
{
	VENTRY *vent;
	int error;

	// zt@CVXe̎QƊJn
	refVfs();

	error = getVent(i_path, O_RDONLY, lookupObj, &vent);
	if (error != NOERR) {
		goto ERR;
	}
	error = vent->fs->stat(vent->din, vent->inodeObj, m_stat);
	releaseVent(vent);
ERR:
	// zt@CVXe̎QƏI
	refEndVfs();
	
	return error;
}

/*
 * pXt@CXe[g擾
 * return : error number
 */
int statPathSymlink(
	const char *i_path,
	struct stat *m_stat)
{
	VENTRY *vent;
	int error;

	// zt@CVXe̎QƊJn
	refVfs();

	error = getVent(i_path, O_RDONLY, lookupObjSymlink, &vent);
	if (error != NOERR) {
		goto ERR;
	}
	error = vent->fs->stat(vent->din, vent->inodeObj, m_stat);
	releaseVent(vent);
ERR:
	// zt@CVXe̎QƏI
	refEndVfs();
	
	return error;
}

/*
 * return : error number
 */
int statVent(
	void *i_vent,
	struct stat *m_stat)
{
	VENTRY *vent = i_vent;
	
	return vent->fs->stat(vent->din, vent->inodeObj, m_stat);
}

/*
 * return : error number
 */
int statfsPath(
	const char *i_path,
	struct statfs *m_statfs)
{
	const char *path;
	const char *parentPath;
	VENTRY *parentVent;
	VENTRY *vent;
	int error;

	// zt@CVXe̎QƊJn
	refVfs();

	//JgfBNg擾B
	vent = getTopVdir(i_path, &path);

	/* fBNgTB */
	error = searchVent(path, vent, &parentPath, &parentVent, &vent);
	switch (error){
	case NOERR:		// zt@Cɂ
		if (vent->mounted == NULL) {
			VENTRY *p;

			for (p = vent; p != rootEnt; p = p->parent) {
				if (p->din == vent->din){
					break;
				}
			}
			memcpy(m_statfs->f_mntonname, p->name, p->len);			/* directory on which mounted */
			m_statfs->f_mntonname[p->len] = '\0';
		}
		else{
			memcpy(m_statfs->f_mntonname, vent->name, vent->len);	/* directory on which mounted */
			m_statfs->f_mntonname[vent->len] = '\0';
		}
		break;
	case -ENOENT:	// zt@Cɖ
		goto ERR;
	default:
		goto ERR;
	}
	
	error = getFsstat(vent, m_statfs);

	// zt@CVXe̎QƏI
	refEndVfs();
ERR:
	return error;
}

/*
 * return : error number
 */
int readlinkPath(
	const char *i_path,
	const size_t bufSize,
	char *readBuf)
{
	VENTRY *vent;
	struct stat stat;
	size_t size;
	int error;

	// zt@CVXe̎QƊJn
	refVfs();

	error = getVent(i_path, O_RDONLY, lookupObjSymlink, &vent);
	if (error != NOERR) {
		goto  ERR;
	}

	/* t@C^CvmF */
	error = vent->fs->stat(vent->din, vent->inodeObj, &stat);
	if (error != 0){
		goto ERR2;
	}
	if ((stat.st_mode & S_IFMT) != S_IFLNK){
		error = -EINVAL;
		goto ERR2;
	}
	
	/* NpXǂ */
	size = (stat.st_size < bufSize)? stat.st_size : bufSize;
	error = vent->fs->read(vent->din, vent->inodeObj, readBuf, size, 0);
ERR2:	
	releaseVent(vent);
ERR:
	// zt@CVXe̎QƏI
	refEndVfs();

	return error;
}

/*
 * return : error number
 */
int chmodPath(
	const char *i_path,
	const mode_t mode)
{
	VENTRY *vent;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = getVent(i_path, O_RDWR, lookupObj, &vent);
	if (error != NOERR) {
		goto ERR;
	}
	error = chmod(vent, mode);
	releaseVent(vent);
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int chmodVent(
	void *i_vent,
	const mode_t mode)
{
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = chmod(i_vent, mode);

	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int chownPath(
	const char *i_path,
	const uid_t owner,
	const gid_t group)
{
	VENTRY *vent;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = getVent(i_path, O_RDWR, lookupObjSymlink, &vent);
	if (error != NOERR) {
		goto ERR;
	}

	if (owner != -1) {
		error = chuid(vent, owner);
		goto ERR2;
	}
	if (group != -1) {
		error = chgrp(vent, group);
	}
ERR2:
	releaseVent(vent);
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int chownVent(
	void *i_vent,
	const uid_t owner,
	const gid_t group)
{
	VENTRY *vent = i_vent;
	int error = NOERR;

	// zt@CVXe̍XVJn
	modifyVfs();

	if (owner != -1) {
		error = chuid(vent, owner);
		goto ERR;
	}
	if (group != -1) {
		error = chgrp(vent, group);
	}
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int utimesPath(
	const char *i_path,
	const struct timeval *times)
{
	VENTRY *vent;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = getVent(i_path, O_RDWR, lookupObj, &vent);
	if (error != NOERR) {
		goto ERR;
	}

	if (times != NULL) {
		error = chutime(vent, times[0].tv_sec, times[1].tv_sec);
	}
	else {
		error = chutime(vent, sys_time(NULL), sys_time(NULL));
	}

	releaseVent(vent);
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int linkPath(
	const char *fromPath,
	const char *toPath)
{
	const char *path;
	const char *parentPath;
	VENTRY *parentVent;
	VENTRY *fromVent;
	VENTRY *vent;
	int error;

	// zt@CVXe̍XVJn
	modifyVfs();

	error = getVent(fromPath, O_RDONLY, lookupObj, &fromVent);
	if (error != NOERR) {
		goto ERR;
	}
	
	//JgfBNg擾B
	vent = getTopVdir(toPath, &path);

	/* toPath̃t@CB */
	error = searchVent(path, vent, &parentPath, &parentVent, &vent);
	switch (error){
	case NOERR:		// zt@Cɂ
		error = -EEXIST;
		break;
	case -ENOENT:	// zt@Cɖ
		if (parentVent->din != fromVent->din) {
			error = -EXDEV;
		}
		else {
			error = parentVent->fs->link(parentVent->din, fromVent->inodeObj, parentVent->inodeObj, parentPath);
		}
		break;
	default:
		break;
	}

	releaseVent(fromVent);
ERR:
	// zt@CVXe̍XVI
	modifyEndVfs();

	return error;
}

/*
 * return : error number
 */
int ioctlVent(
	void *i_vent,
	const int cmd,
	caddr_t parm,
	int fflag)
{
	VENTRY *vent = i_vent;

	return vent->fs->ioctl(vent->din, vent->inodeObj, cmd, parm, fflag);
}

/*
 * return : zGg[
 */
void *getChdirVentPath(
	const char *i_path,
	int *o_error)
{
	VENTRY *vent;
	VENTRY *mountedVent;

	*o_error = openPath(i_path, O_RDWR, (void**) &vent);
	if (*o_error != NOERR) {
		return NULL;
	}
	mountedVent = getMountedVent(vent);
	if (mountedVent != vent) {
		closeVent(vent);
		linkVent(mountedVent);
	}
	if (mountedVent->type != S_IFDIR){
		closeVent(mountedVent);
		*o_error = -ENOTDIR;
		return NULL;
	}

	return mountedVent;
}

/*
 * return : zGg[
 */
void *getChdirVent(
	void *i_vent,
	int *o_error)
{
	VENTRY *vent;

	vent = getMountedVent(i_vent);
	if (vent->type != S_IFDIR) {
		*o_error = -ENOTDIR;
		return NULL;
	}
	linkVent(vent);

	*o_error = NOERR;
	return vent;
}

/*
 * retern : poll condition
 */
int pollVent(
	void *i_vent,
	int events)			// events
{
	VENTRY *vent = i_vent;

	return vent->fs->poll(vent->inodeObj, events);
}

/*
 * pX[g܂şăobt@ɃRs[B
 * return : Rs[TCY
 */
int copyRootPath(
	const void *vent,		// XgGg[
	const size_t size,		// Rs[őTCY
	char *m_path)			// Rs[obt@
{
	int len;

	// zt@CVXe̎QƊJn
	refVfs();

	len = copyPathBack(vent, size, isRoot, m_path);

	// zt@CVXe̎QƏI
	refEndVfs();
	
	return len;
}

/*
 * }EgGg[B
 * return : struct statfs֊i[ or }Egistruct statfs=NULLj
 */
int searchMount(struct statfs *m_statfs, const int i_count)
{
	struct statfs *statfs = m_statfs;
	int count = i_count;
	int rest;

	if (count == 0) {
		return 0;
	}
	
	// zt@CVXe̎QƊJn
	refVfs();

	if (statfs != NULL){
		// roott@CVXe
		getFsstat(rootEnt, statfs);
		statfs += 1;
	}
	count -= 1;
	rest = searchMountDir(rootEnt->child, statfs, count);

	// zt@CVXe̎QƏI
	refEndVfs();

	return rest + 1;
}

/*
 * zGg[t@C^Cv擾
 * return : t@C^Cv
 */
int getTypeVent(void *i_vent)
{
	VENTRY *vent = i_vent;

	return vent->type;
}

/*
 * zGg[foCXinode擾
 * return : t@C^Cv
 */
int getDinVent(void *i_vent)
{
	VENTRY *vent = i_vent;

	return vent->din;
}

/*
 * zGg[inodeIuWFNg擾
 * return : t@C^Cv
 */
void *getInodeObjVent(void *i_vent)
{
	VENTRY *vent = i_vent;

	return vent->inodeObj;
}

/*
 * [gzGg[擾B
 * return : [gzGg[
 */
void *getRootEnt()
{
	return rootEnt;
}

//------------------ zt@CVXȅ -------------------

void initVfs()
{
	/* rootGg[̏B */
	rootEnt->fs			= NULL;
//	rootEnt->inodeObj	= NULL;
	rootEnt->inodeObj	= (void*) 1;	//[gt@CVXe}Eg searchPth()ŃG[ɂȂ߂̉}u
	rootEnt->next		= rootEnt;
	rootEnt->prev		= rootEnt;
	rootEnt->parent		= rootEnt;
	rootEnt->child		= NULL;
	rootEnt->mount		= NULL;
	rootEnt->linkCount	= 2;		// ftHgJgfBNgȂ̂łQݒ肷B
	rootEnt->din		= 0;
	rootEnt->type		= S_IFDIR;
	rootEnt->len		= 1;
	rootEnt->name[0]	= '/';
}
